# This file belongs to Happy Camel.
#
# Copyright (C) 2006, 2007 Pieter Edelman (p _dot_ edelman AT gmx _dot_ net)
#
# Happy Camel is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation; version 2 only.
#
# Happy Camel is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

""" Module to handle efficient communction with the Geonames.org service. """

import urllib, urlparse, xml.dom.minidom, math, bisect, copy

class Point:
  """ Really simple container for storing a geographical point. It has a
      latitude, longitude, name, country_code, and country_name. The first two
      need to be specified during initialization, the othe three are initially
      None. """
  
  def __init__(self, lat, lon):
    """ Initialize with latitude and longitude. """
    self.lat          = lat
    self.lon          = lon
    self.name         = None
    self.country_code = None
    self.country_name = None

class GeonamesReader(urllib.FancyURLopener):
  """ The class for communicating with geonames.org. """
  
  # Let geonames know who we are
  version = "Happy Camel/v1-beta3" # FIXME: This should be read from central file

  # The base URL for all Geonames webservices.
  base_url = "ws.geonames.org"
  
  def __init__(self, lang = "en", proxies = None, **x509):
    """ Initialize the Geonames reader. Language is an optional two-letter
        language code which specified the language of the returned results. """

    self.lang = lang
    urllib.FancyURLopener.__init__(self, proxies, **x509)
    
  def query(self, module, **parameters):
    """ Query the Geonames database with the specified module and parameters
        and return a digested XML result. """
    
    # Inject the language parameter if needed
    if ("lang" not in parameters):
      parameters["lang"] = self.lang
      
    # Construct the parameter string
    param_string = ""
    for param_name in parameters:
      param_string += urllib.quote(param_name)
      param_string += "="
      param_string += urllib.quote(str(parameters[param_name]))
      param_string += "&"
    param_string = param_string[:-1]

    # Make the query
    url = urlparse.urlunparse(["http", self.base_url, module, "", param_string, ""])
    response = self.open(url).read()

    # Parse the (REST)-response
    response = xml.dom.minidom.parseString(response)
    
    return response

class Geonames:
  """ Class for querying the online Geonames service (geonames.org). """
  
  # The average Earth radius in meters
  earth_radius = 6371000
  
  def __init__(self, lang = "en", country_cache = 50):
    """ lang may be a two-letter language code, and specifies the language
        the results are returned in. The country_cache specifies the distance
        in meters over which the results for the country lookups are cached. """

    # We keep a cache of coordinates and the country they are in. If a point is
    # within a certain distance form a point in the list (designated by 
    # cache_dist), it is assumed the new point is also in that country. This
    # can reduce the number of calls to geonames.org
    self.cache_dist   = country_cache
    self.cache_points = []

    # The communication with geonames.org
    self.gn_reader = GeonamesReader(lang)
    
  def getCountry(self, lat, lon):
    """ Return a dict with country_code and country_name for a given latitude
        and longitude, or False if nothing can be found. """

    ret = {}
    
    # Try if the point is inside the cache distance of a previous point
    for coord in self.cache_points:
      if (self.__getDifference__(lat, lon, coord.lat, coord.lon) < self.cache_dist):
        ret["country_code"] = coord.country_code
        ret["country_name"] = coord.country_name
        break
        
    # If we didn't get anything from the cache, query the Geonames database
    if (len(ret) == 0):
      # Add the point to the cache
      self.cache_points.append(Point(lat, lon))
      
      # Query the Geonames database
      response = self.gn_reader.query("countrycode", lat = lat, lng = lon, type = "xml")
      
      # Analyze the response
      if response:
        try:
          # Get country code and name
          country_name_nodes = response.getElementsByTagName("geonames")[0].getElementsByTagName("country")[0].getElementsByTagName("countryName")[0].childNodes
          country_code_nodes = response.getElementsByTagName("geonames")[0].getElementsByTagName("country")[0].getElementsByTagName("countryCode")[0].childNodes
          if (len(country_name_nodes) > 0):
            country_name = country_name_nodes[0].wholeText
          else:
            country_name = None
          if (len(country_code_nodes) > 0):
            country_code = country_code_nodes[0].wholeText
          else:
            country_code = None
          
          # Set the return value
          ret["country_code"] = country_code
          ret["country_name"] = country_name
          
          # Update the cache (if no country was found, the point is still in the
          # cache with country values None.
          self.cache_points[-1].country_code = country_code
          self.cache_points[-1].country_name = country_name
        except AttributeError:
          # If we didn't find anything, we should return False
          ret = False

    return ret
    
  def getPlace(self, lat, lon, radius = 10):
    """ Find the nearest sub-place, place and country within the specified
        radius (in kilometer). The return value is a dict with keys sub_place,
        place, country_name and country_code. If nothing can be found, response
        is False. """
    
    ret = {}
    
    # Query geonames.org
    response = self.gn_reader.query("findNearbyPlaceName", lat = lat, lng = lon, radius = radius)

    # Analyze the response
    if response:
      # Loop over the responses
      for location in response.childNodes[0].getElementsByTagName("geoname"):
        location_type = location.getElementsByTagName("fcode")[0].childNodes[0].wholeText
        
        # Try to set the country name on the first occasion (geonames.org
        # returns the most nearby results first.
        if ("country_name" not in ret):
          if len(location.getElementsByTagName("countryName")):
            ret["country_name"] = location.getElementsByTagName("countryName")[0].childNodes[0].wholeText
            ret["country_code"] = location.getElementsByTagName("countryCode")[0].childNodes[0].wholeText
            
        if (location_type == "PPLX") and ("sub_place" not in ret):
          # We found a sub location
          ret["sub_place"] = location.getElementsByTagName("name")[0].childNodes[0].wholeText
        elif (location_type in ["PPL", "PPLA", "PPLC", "PPLG"]) and ("place" not in ret):
          # We found a location
          ret["place"] = location.getElementsByTagName("name")[0].childNodes[0].wholeText
          # We should stop processing here, because we don't want a sub location
          # to be further away thant the main location 
          break

    return ret

  def getPlaces(self, coordinates, radius = 10, query_country = True, query_place = True, query_sub_place = True):
    """ Return a list of place objects when fed with a list of lattitudes and 
        longitudes. The optional radius parameter specifies the maximum distance
        the place may be away from the specified coordinate (up to 25 km). The
        optional booleans query_country, query_place and query_sub_place specify
        whether to query the country, place (populated place, usually a village
        or a city) and sub place (area within a place). """
    
    # Do sanity check for radius
    if (radius > 25):
      raise ValueError, "Please select a radius of maximally 25 kilometers."
      
    # Create a list of coordinates and a list of return dicts
    points = []
    ret    = []
    for coordinate in coordinates:
      points.append(Point(coordinate[0], coordinate[1]))
      ret.append({"country_name": None, "country_code": None, "place": None, "sub_place": None})

    # If we need to find places and/or sub places, create the lists of all
    # places within range
    if (query_place) or (query_sub_place):
      # Get the centers of each cluster from where we should make a call to
      # geonames
      centers = self.__getClusterCenters__(points, radius)
  
      # We buffer all the places and sub_places we know of so we can search in
      # them efficiently.
      places     = []
      sub_places = []

      # Make a request for each cluster center
      for center in centers:
        # The radius should be twice that of the user requested radius, because
        # then even places on the edge get their full radius search. Geonamses.org
        # by default returns 10 results, so we ask for 1000 (the maximum allowed)
        # to make sure that all places in the radius are returned (only places
        # within the radius are returned).
        response = self.gn_reader.query("findNearbyPlaceName", lat = center.lat, lng = center.lon, radius = 2 * radius, maxRows = 1000)
        
        # Parse the response
        if response:
          for location in response.childNodes[0].getElementsByTagName("geoname"):
            # Retreive latitude and longitude and create a point object
            lat = float(location.getElementsByTagName("lat")[0].childNodes[0].wholeText)
            lon = float(location.getElementsByTagName("lng")[0].childNodes[0].wholeText)
            point = Point(lat, lon)
            
            # Retrieve country info and store it in the point object
            point.country_name = location.getElementsByTagName("countryName")[0].childNodes[0].wholeText
            point.country_code = location.getElementsByTagName("countryCode")[0].childNodes[0].wholeText
            point.name = location.getElementsByTagName("name")[0].childNodes[0].wholeText
            
            # Retrieve location type, and store the point in the appropriate list
            # based on that
            location_type = location.getElementsByTagName("fcode")[0].childNodes[0].wholeText
            places_list = None
            if (location_type in ["PPL", "PPLA", "PPLC", "PPLG"]) and (query_place):
              # We found a location
              places_list = places
            elif (location_type == "PPLX") and (query_sub_place):
              # We found a sub location
              places_list = sub_places
            
            # Add the point if it doesn't already occur in the list
            if (places_list != None):
              add_point = True
              for place in places_list:
                if (point.name == place.name):
                  add_point = False
              if (add_point):
                places_list.append(point)
              
      # Sort the lists of places and sub places by lattitude. This is needed for
      # the __getNearest__ method
      places.sort(self.sortByLat)
      sub_places.sort(self.sortByLat)

    # Get the place information for each point
    for point_num in range(len(points)):
      # Get place and sub_place
      if (query_place):
        place = self.__getNearest__(points[point_num].lat, points[point_num].lon, radius, places)
      else:
        place = None
        
      if (query_sub_place):
        sub_place = self.__getNearest__(points[point_num].lat, points[point_num].lon, radius, sub_places)
      else:
        sub_place = None
      
      # country_source is the place where we should obtain the country info
      # from. We will use sub place if it is neare than the place or if there is
      # no place.
      country_source = None
      
      # Set the required info to the return dict
      if (place):
        ret[point_num]["place"] = place.name
        country_source = place
        
      if (sub_place) and ((not place) or (sub_place.dist <= place.dist)):
        ret[point_num]["sub_place"] = sub_place.name
        country_source = sub_place
      
      if (country_source) and (query_country):
        ret[point_num]["country_name"] = country_source.country_name
        ret[point_num]["country_code"] = country_source.country_code

      # If we couldn't find a nearby place/sub_place (or didn't look for it), we
      # don't have a country so we have to query for it separately.
      if (not ret[point_num]["country_name"]) and (query_country):
        country_info = self.getCountry(points[point_num].lat, points[point_num].lon)
        if (country_info):
          ret[point_num]["country_name"] = country_info["country_name"]
          ret[point_num]["country_code"] = country_info["country_code"]
    
    if (len(ret) == 0):
      ret = False

    return ret

  def __getDifference__(self, lat1, lon1, lat2, lon2):
    """ Return the distance in meters between the coordinates defined by
        (lat1, lon2) and (lat2, lon2). """
        
    # There are different methods to calculate the exact position on Earth. We
    # use the most simple one, based on the spherical law of cosines. The
    # flattening of the Earth isn't taken into account here, but the precision
    # is enough for our purpose, about 0.5 percent maximum.
    
    # Convert to radians
    latr1 = math.radians(lat1)
    latr2 = math.radians(lat2)
    lonr1 = math.radians(lon1)
    lonr2 = math.radians(lon2)
    
    # Do the calculation
    d = math.sin(latr1) * math.sin(latr2) + math.cos(latr1) * math.cos(latr2) * math.cos(lonr2 - lonr1)
    
    # In some cases, a small floating point round-off error can crash math.acos,
    # so we have to cap it off at |1|
    if (d > 1): d = 1
    elif (d < -1): d = -1
    
    # Get the cosine
    diff = math.acos(d) * self.earth_radius
    
    return diff

  def sortByLat(self, point1, point2):
    """ Callback method to sort places by their latitude. """
    
    if (point1.lat == point2.lat):
      return 0
    elif (point1.lat > point2.lat):
      return 1
    else:
      return -1
    
  def __getClusterCenters__(self, points, radius):
    """ Get the minimum number of points which need to be queried to cover the
        complete list of specified points. Warning: this method may produce
        inconsequent results for the same input, although it wil always return
        a comlete coverage. """
        
    # We will use the Quality Threshold (QT) clustering algorithm to find the
    # center points for the clusters. This method can work with clusters of a
    # fixed size. Unfortunately, it doesn't yield the lowest number of clusters
    # possible, but I'm not aware of an algorithm that can.
    # TODO: Research boundary limitations; a point on the side always needs to 
    # be in at least one cluster, so that may be a good starting point.
    
    # Store all the cluster centers
    centers = []
    
    temp_points = copy.copy(points)
    
    while (len(temp_points)):
      # Build a candidate cluster for each point
      candidates = {}
      for base in temp_points:
        candidates[base] = []
        for place in temp_points:
          # Include one self
          if (base == place):
            candidates[base].append(place)
          # Include any place that falls within the radius
          else:
            if (self.__getDifference__(base.lat, base.lon, place.lat, place.lon) < (radius * 1000)):
              candidates[base].append(place)

      # Find the cluster with the most places (the first true cluster)
      max_places   = 0
      true_cluster = None
      for cluster in candidates.keys():
        if (len(candidates[cluster]) > max_places):
          max_places = len(candidates[cluster])
          true_cluster = cluster
      
      # Remove the places in the cluster from the list
      for place in candidates[true_cluster]:
        temp_points.remove(place)
        
      # Save the true cluster coordinates
      centers.append(true_cluster)
      
    return centers
      
  def __getNearest__(self, lat, lon, radius, places):
    if (len(places)):
      # Do a binary search for the latitude position of our point
      lat_places = []
      for place in places:
        lat_places.append(place.lat)
      index = bisect.bisect(lat_places, lat)
      
      # Determine the size to its latitude-neighbors
      ldist = rdist = None
      if (index > 0):
        ldist = self.__getDifference__(lat, lon, places[index - 1].lat, places[index - 1].lon)
      if (index < len(lat_places)):
        rdist = self.__getDifference__(lat, lon, places[index].lat, places[index].lon)
        
      # Choose the smallest size
      if (ldist == None) or ((rdist != None) and (rdist < ldist)):
        smallest_dist = rdist
        nearest_place = places[index]
      else:
        smallest_dist = ldist
        nearest_place = places[index - 1]
      
      # Create a set of points that along the longitude that fall into the
      # smallest size.
      # Divide the distance by the Earth radius to obtain the angle in radians,
      # and convert this to degrees
      angle = math.degrees(smallest_dist / self.earth_radius)
      # FIXME: You migth run into trouble at the poles
      min_lat = lat - angle
      max_lat = lat + angle
  
      valid_places = []
      for place in places:
        if (place.lat >= min_lat) and (place.lat <= max_lat):
          valid_places.append(place)
      
      # Determine the distance to each of these points and choose the smallest
      # distance
      for place in valid_places:
        dist = self.__getDifference__(lat, lon, place.lat, place.lon)
        if (dist < smallest_dist):
          smallest_dist = dist
          nearest_place = place
      
      if (smallest_dist < (radius * 1000)):
        ret = copy.copy(nearest_place)
        ret.dist = smallest_dist
        return ret

    return None

