# This file belongs to Happy Camel.
#
# Copyright (C) 2006-2008 Pieter Edelman (p _dot_ edelman AT gmx _dot_ net)
#
# Happy Camel is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation; version 2 only.
#
# Happy Camel is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

""" Module to create Google Earth files with photo thumbnails and tracks. """

import os, zipfile, threading, Queue, urllib, StringIO

class KML:
  """ Class to create a KML file, the XML file for Google Earth sans the
      binary data like photo thumbnails. """

  def __init__(self, gpx_files, photo_list, thumbnail_names, logger,
               title = "Photo report", track_color = "ffffffff",
               track_width = 1.0,  between_tracks_color = None,
               between_tracks_width = None, photos_clickable = False):
    """ Initialize a new KML file. gpx_files is a list of GPXFile objects which
        should be drawn in the file. photo_list is a PhotoList object of
        photo's whose thumbnail should be included. thumbnail_names is an
        associated dict of file names of all the thumbnails in the zip file.
        title is a string determining the title of the report shown in Google
        Earth.
        This class shouldn't be called directly, it acts as a helper for the KMZ
        class. """

    self.logger           = logger
    self.gpx_files        = gpx_files
    self.photo_list       = photo_list
    self.thumbnail_names  = thumbnail_names
    self.title            = title
    self.photos_clickable = photos_clickable

    # Check if the specified track colors and widths are ok
    self.track_color          = self.__testColorFormat__(track_color)
    self.between_tracks_color = self.__testColorFormat__(between_tracks_color)
    if (track_width < 0.0):
      self.logger.error("Track width is too small, setting it to 0.0")
      track_width = 0.0
    self.track_width = track_width
    if (between_tracks_width < 0.0):
      self.logger.error("Track width is too small, setting it to 0.0")
      between_tracks_width = 0.0
    self.between_tracks_width = between_tracks_width

  def writeKML(self, out_fp, write_photos = True, write_waypoints = True,
               write_track = True, write_between_tracks = True):
    """ Write a KML file to the specified file pointer out_fp. The arguments
        specify which parts of the file should be included. If write_photos is
        True, thumbnails should be included. If write_waypoints is True,
        waypoints should be included. If write_track is True, the tracks should
        be included. If write_between_tracks is True, lines should be drawn
        between tracks."""
    out_fp.write('<?xml version="1.0" encoding="UTF-8"?>\n')
    out_fp.write('<Document xmlns:xlink="http://www.w3/org/1999/xlink">\n')
    out_fp.write('  <name>%s</name>\n' % self.title)

    # Get the waypoints and the tracks
    waypoints = []
    tracks    = []
    for gpx_file in self.gpx_files:
      track_waypoints, track_sub_tracks = gpx_file.getWaypointsAndSimpleTrack()
      waypoints += track_waypoints
      tracks.append(track_sub_tracks)

    if (write_photos):
      self.logger.debug("Writing photo folder")
      self.writePhotos(out_fp)

    if (write_waypoints):
      self.logger.debug("Writing waypoints folder")
      self.writeWaypoints(out_fp, waypoints)

    if (write_track) or (write_between_tracks):
      self.logger.debug("Writing track folder")
      self.writeTracks(tracks, out_fp, write_track, write_between_tracks)

    out_fp.write('</Document>\n')

  def writePhotos(self, out_fp):
    """ Write the photo references with comments and stuff to the file pointer
        out_fp . """

    # Write the folder
    self.logger.debug("Writing the photo folder in the KML file")
    out_fp.write('<Style id="photo_mark">\n')
    out_fp.write('  <IconStyle>\n')
    out_fp.write('  <scale>0.5</scale>\n')
    out_fp.write('    <Icon>\n')
    out_fp.write('      <href>root://icons/palette-4.png</href>\n')
    out_fp.write('      <x>192</x>\n')
    out_fp.write('      <y>64</y>\n')
    out_fp.write('      <w>32</w>\n')
    out_fp.write('      <h>32</h>\n')
    out_fp.write('    </Icon>\n')
    out_fp.write('  </IconStyle>\n')
    out_fp.write('  <LabelStyle>\n')
    out_fp.write('    <scale>0.2</scale>\n')
    out_fp.write('  </LabelStyle>\n')
    out_fp.write('</Style>\n')
    out_fp.write('<Folder>\n')
    out_fp.write('  <name>Images</name>\n')
    # Write each thumbnail as a placemark
    for photo in self.photo_list:
      if (photo in self.thumbnail_names): # Only process if there's a thumbnail
        location = photo.getLocation()
        out_fp.write('  <Placemark>\n')
        out_fp.write('    <name>%s</name>\n' % photo.getName())
        out_fp.write('    <description><![CDATA[')
        if (self.photos_clickable):
          out_fp.write('<a href="%s">' % photo.getURL())
        out_fp.write('<img src="images/%s" />' % self.thumbnail_names[photo])
        if (self.photos_clickable):
          out_fp.write("</a>")
        comment = photo.getComment()
        if (comment):
          out_fp.write('<p>%s</p>' % comment)
        out_fp.write(']]></description>\n')
        if (comment):
          if (len(comment) > 100):
            out_fp.write('    <Snippet><![CDATA[%s...]]></Snippet>\n' % comment[0:30])
          else:
            out_fp.write('    <Snippet><![CDATA[%s]]></Snippet>\n' % comment)
        else:
          out_fp.write('    <Snippet>%s</Snippet>\n' % photo.getName())
        out_fp.write('    <styleUrl>#photo_mark</styleUrl>\n')
        out_fp.write('    <Point>\n')
        if (location[2]):
          out_fp.write('      <altitudeMode>clampToGround</altitudeMode>\n')
          out_fp.write('      <coordinates>%.5f,%.5f,%.3f</coordinates>\n' % (location[1], location[0], location[2]))
        else:
          out_fp.write('      <altitudeMode>clampedToGround</altitudeMode>\n')
          out_fp.write('      <coordinates>%.5f,%.5f,0</coordinates>\n' % (location[1], location[0]))
        out_fp.write('    </Point>\n')
        out_fp.write('  </Placemark>\n')
      else:
        self.logger.error("No thumbnail for photo %s - skipping." % photo.getName())
    out_fp.write('</Folder>\n')

  def writeWaypoints(self, out_fp, waypoints):
    """ Write the specified waypoints to the file pointer out_fp. waypoints
        should be a list or tuple with waypoints. Each waypoint should be
        specified as:
          [latitude, longitude, altitude, name]
        where altitude may be None. """

    if (len(waypoints)):
      # The style for waypoints
      out_fp.write('<Style id="waypoint_style">\n')
      out_fp.write('  <IconStyle>\n')
      out_fp.write('  <scale>0.7</scale>\n')
      out_fp.write('    <Icon>\n')
      out_fp.write('      <href>root://icons/palette-4.png</href>\n')
      out_fp.write('      <x>160</x>\n')
      out_fp.write('      <y>128</y>\n')
      out_fp.write('      <w>32</w>\n')
      out_fp.write('      <h>32</h>\n')
      out_fp.write('    </Icon>\n')
      out_fp.write('  </IconStyle>\n')
      out_fp.write('</Style>\n')

      # Write the waypoints folder
      out_fp.write('<Folder>\n')
      out_fp.write('  <name>Waypoints</name>\n')

      # Write the actual waypoints
      for waypoint in waypoints:
        out_fp.write('  <Placemark id="%s">\n' % waypoint[3])
        out_fp.write('  <styleUrl>waypoint_style</styleUrl>\n')
        out_fp.write('  <Point>\n')
        out_fp.write('    <coordinates>%.5f,%.5f,%.3f</coordinates>\n' % (waypoint[1], waypoint[0], (waypoint[2] or 0)))
        out_fp.write('  </Point>\n')
        out_fp.write('  <name>%s</name>\n' % waypoint[3])
        out_fp.write('  </Placemark>\n')
      # Close the folder
      out_fp.write('</Folder>\n')

  def writeTracks(self, tracks, out_fp, write_track, write_between_tracks):
    """ Write the tracks to the file handle out_fp. write_track and
        write_between_tracks are booleans that specify whether to write the
        track and to draw lines in between tracks. """

    if (len(tracks)):
      # Write the track styles
      if (write_track):
        out_fp.write('<Style id="track">\n')
        out_fp.write('  <LineStyle>\n')
        out_fp.write('    <color>%s</color>\n' % self.track_color)
        out_fp.write('    <width>%f</width>\n' % self.track_width)
        out_fp.write('  </LineStyle>\n')
        out_fp.write('</Style>\n')
      if (write_between_tracks):
        out_fp.write('<Style id="betweentracks">\n')
        out_fp.write('  <LineStyle>\n')
        out_fp.write('    <color>%s</color>\n' % self.between_tracks_color)
        out_fp.write('    <width>%f</width>\n' % self.between_tracks_width)
        out_fp.write('  </LineStyle>\n')
        out_fp.write('</Style>\n')

      # Write the track folder
      out_fp.write('<Folder>\n')
      out_fp.write('  <name>Tracks</name>\n')

      # Write all the subtracks
      end_point = False
      for track in tracks:
        for sub_track in track:
          if (end_point) and (write_between_tracks):
            self.__writeTrack__([end_point, sub_track[0]], out_fp, "#betweentracks")
          if (write_track):
            self.__writeTrack__(sub_track, out_fp, "#track")
          end_point = sub_track[-1]

      # Close the folder
      out_fp.write('</Folder>\n')

  def __writeTrack__(self, track, out_fp, style):
    """ Do the low-level track writing with the defined style name. """
    out_fp.write('  <Placemark>\n')
    out_fp.write('    <name>Track</name>\n')
    out_fp.write('    <styleUrl>%s</styleUrl>\n' % style)
    out_fp.write('    <MultiGeometry>\n')
    out_fp.write('      <LineString>\n')
    out_fp.write('        <coordinates>\n')
    # Write all the coordinates
    for pos in track:
      out_fp.write('%.5f,%.5f,%.3f ' % (pos[1], pos[0], pos[2]))
    out_fp.write('\n        </coordinates>\n')
    out_fp.write('      </LineString>\n')
    out_fp.write('    </MultiGeometry>\n')
    out_fp.write('  </Placemark>\n')

  def __testColorFormat__(self, color):
    """ Tests whether the specified color is properly formatted, and returns a
        sensible alternative if not. """

    # Try if we have a hexadecimal number
    try:
      int('0x%s' % color, 16)
    except ValueError:
      self.logger.error("Track color %s is not given in the right format, setting it to ffffffff" % color)
      return("ffffffff")

    # Check for the proper lenght (eight chars), and if not, break it off or
    # create all f's
    if (len(color) > 8):
      self.logger.error("Track color %s is too long, clipping it to %s" % color[:8])
      return color[:8]
    elif (len(color) < 8):
      return "ffffffff"

    return color


class KMZ:
  """ Class for creating Google Earth archives, that is a KML file with all
      associated binary data like photo thumbnails. """

  # We only can keep a certain number of thumbnails in memory, if this is
  # surpassed, we first make to sure we flush it all to disk before we can
  # continue.
  NUM_THUMBNAILS = 25

  def __init__(self, file_name, logger, gpx_files = None, photo_list = None,
               sort_order = "D", thumbnail_size = 200, title = "Photo report",
               track_color = "ffffffff", track_width = 1.0,
               between_tracks_color = None, between_tracks_width = None,
               photos_clickable = False):
    """ Create a new KMZ file with the disk path file_name. gpx_files is a list
        of GPXFile objects that need to be showed in the file. photo_list is a
        PhotoList object of photo's whose thumbnail should be included.
        sort_order is a string which specifies in which order the photo's are
        presented ("D" is date and time, "A" is latitude, "O" is longitude.
        Lowercase means reverse sort. Multiple tokens may be specified in order
        of precedence).
        thumbnail_size is the number of pixels the long size of each thumbnail
        should have.
        title is the title given to the report in Google Earth.
        track_color and between_tracks_color specify the color to give the
        tracks and the lines in between the tracks, respectively. They are
        specified as 4 2-digit hexadecimal numbers specifying opacity, blue,
        green and red.
        track_width and between_track_width are floats specifying the width in
        pixels of the tracks and the lines in between the tracks, respectively.
        If photos_clickable is true, the thumbnails are hyperlinked to the path
        on disk of the original photo. """

    self.logger = logger

    name_parts = os.path.splitext(file_name)
    self.base_name = file_name
    if (len(name_parts) > 1):
      if (name_parts[1].lower() == ".kmz"):
        self.base_name = name_parts[0]

    self.gpx_files            = gpx_files
    self.photo_list           = photo_list
    self.thumbnail_size       = thumbnail_size
    self.title                = title
    self.track_width          = track_width
    self.track_color          = track_color
    self.between_tracks_width = between_tracks_width
    self.between_tracks_color = between_tracks_color
    self.photos_clickable     = photos_clickable

    # Sort the photo's
    for index in range(len(sort_order)):
      action = sort_order[-index]
      if (action == "D"):
        self.photo_list.sortByDate()
      elif (action == "d"):
        self.photo_list.sortByDate(reverse = True)
      if (action == "A"):
        self.photo_list.sortByLat()
      elif (action == "a"):
        self.photo_list.sortByLat(reverse = True)
      if (action == "O"):
        self.photo_list.sortByLon()
      elif (action == "o"):
        self.photo_list.sortByLon(reverse = True)

  def __processThumbnails__(self, queue, zip_file, file_names):
    """ Process a que of thumbnails by adding them to the zipfile. The queue
        should be a Queue.Queue instance, containg lists of the photo object and
        the associated JPEG-encoded thumbnail data.
        zip_file is an open zipfile.ZipFile object.
        file_names is a dict to store the generated thumbnail file names in. """

    # Process the number of thumbnails in the photo list.
    for counter in range(len(self.photo_list)):
      # Wait until data becomes available
      photo, data = queue.get(block = True)
      
      # Process it
      if (data):
        
        # Find a good file name, based on the photo name
        fn_base = fn_base_raw = os.path.splitext(photo.getName())[0]
        
        # Replace spaces with underscores
        fn_base = fn_base.replace(" ", "_")
        
        # Replace nasty characters with other nasty, but manageable characters
        fn_base = urllib.quote(fn_base)
        
        # Find a unique name
        addition = 0
        while (("%s.jpg" % fn_base) in file_names.values()):
          fn_base = "%s_%2d" % (fn_base_raw, addition)
          addition += 1
          
        # Save the file name
        file_names[photo] = "%s.jpg" % fn_base
        
        # Write the file
        zip_file.writestr("images/%s.jpg" % fn_base, data)
      else:
        self.logger.error("Couldn't make a thumbnail of image %s." % photo.getName())
        
  def writeFile(self, write_photos = False, write_waypoints = False,
                write_track = False, write_between_tracks = False):
    """ Write the KMZ file. The arguments specify which parts of the file should
        be included. If write_photos is True, thumbnails should be included.
        If write_waypoints is True, waypoints should be included. If write_track
        is True, the tracks should be included. If write_between_tracks is True,
        lines should be drawn between tracks. """

    # Open the destination KMZ file as a zip file
    out_file = zipfile.ZipFile(self.base_name + ".kmz", "w")

    # Process the thumbnails
    if (write_photos):
      # Create a queue where we store the thumbnail data
      tn_queue = Queue.Queue(self.NUM_THUMBNAILS)

      # Start a threads to fill the queue and to process the queue. Calling the
      # PhotoLists getThumbnailData() method starts multiple threads in itselt,
      # but we put it here in a separate thread here as well because at the same
      # time we need to start the processing thread.
      get_thread = threading.Thread(target = self.photo_list.getThumbnailData,
                                    args = [self.thumbnail_size, tn_queue])
      file_names = {}
      process_thread = threading.Thread(target = self.__processThumbnails__,
                                        args = [tn_queue, out_file, file_names])

      get_thread.start()
      process_thread.start()
      
      # Now block until all the thumbnails have been put into the queue.
      get_thread.join()
      
      # Wait until the consuming thread finishes.
      process_thread.join()
      # TODO: Write only references to thumbnails that are actually fetched
      
    # Make the KML file
    kml = KML(self.gpx_files, self.photo_list, file_names, self.logger,
              self.title, self.track_color, self.track_width,
              self.between_tracks_color, self.between_tracks_width,
              self.photos_clickable)
    kml_fp = StringIO.StringIO()
    kml.writeKML(kml_fp, write_photos, write_waypoints, write_track, write_between_tracks)

    # Write the KML file to the ZIP archive
    out_file.writestr("base.kml", kml_fp.getvalue())

    # Close everything
    kml_fp.close()
    out_file.close()
